#!/usr/bin/env python

# Copyright (c) 2014, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


"""Network module contains objects that exist in the 'Network' tab in the firewall GUI"""

# import modules
import re
import logging
import xml.etree.ElementTree as ET
import pandevice
from pandevice.base import PanObject, Root, MEMBER, ENTRY
from pandevice.base import VarPath as Var
from pandevice import getlogger, string_or_list
from pandevice import device
from pandevice.base import VersionedPanObject
from pandevice.base import VersionedParamPath
from pandevice.base import VsysOperations

# import other parts of this pandevice package
import pandevice.errors as err

logger = getlogger(__name__)


def interface(name, *args, **kwargs):
    """Interface object factory

    Creates an interface object of type determined by the name of the interface.

    Args:
        name (str): Name of the interface to create (eg. ethernet1/1.5)
        mode (str): Mode of the interface.
            Possible values: layer3, layer2, virtual-wire, tap, ha, aggregate-group.
            Default: None

    Keyword Args:
        tag (int): Tag for the interface, aka vlan id

    Returns:
        Interface: An instantiated subclass of :class:`pandevice.network.Interface`

    """
    name = str(name)
    if name.startswith("ethernet") and '.' not in name:
        return EthernetInterface(name, *args, **kwargs)
    elif name.startswith("ae") and '.' not in name:
        return AggregateInterface(name, *args, **kwargs)
    elif name.startswith("ethernet") or name.startswith("ae"):
        # Subinterface
        # Get mode from args
        args = string_or_list(args)
        if len(args) > 0:
            mode = args[0]
            del args[0]
        else:
            mode = kwargs.pop("mode", None)
        # Get tag from kwargs
        tag = kwargs.get("tag", None)
        if tag is None:
            # Determine tag from name
            tag = name.split(".")[-1]
            kwargs["tag"] = tag
        if mode == "layer3":
            return Layer3Subinterface(name, *args, **kwargs)
        elif mode == "layer2":
            return Layer2Subinterface(name, *args, **kwargs)
        else:
            return AbstractSubinterface(name, *args, **kwargs)
    elif name.startswith("vlan"):
        return VlanInterface(name, *args, **kwargs)
    elif name.startswith("loopback"):
        return LoopbackInterface(name, *args, **kwargs)
    elif name.startswith("tunnel"):
        return TunnelInterface(name, *args, **kwargs)
    else:
        raise err.PanDeviceError("Can't identify interface type from name: %s" % name)


class Zone(VersionedPanObject):
    """Security zone

    Args:
        name (str): Name of the zone
        mode (str): The mode of the security zone. Must match the mode of the interface.
            Possible values: tap, virtual-wire, layer2, layer3, external
        interface (list): List of interface names or instantiated subclasses
            of :class:`pandevice.network.Interface`.
        zone_profile (str): Zone protection profile
        log_setting (str): Log forwarding setting
        enable_user_identification (bool): If user identification is enabled
        include_acl (list/str): User identification ACL include list
        exclude_acl (list/str): User identification ACL exclude list

    """
    ROOT = Root.VSYS
    SUFFIX = ENTRY

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/zone')

        # params
        params = []

        params.append(VersionedParamPath(
            'mode', default='layer3', path='network/{mode}',
            values=['tap', 'virtual-wire', 'layer2', 'layer3', 'external']))
        params.append(VersionedParamPath(
            'interface', path='network/{mode}', vartype='member'))
        params.append(VersionedParamPath(
            'zone_profile', path='network/zone-protection-profile'))
        params.append(VersionedParamPath(
            'log_setting', path='network/log-setting'))
        params.append(VersionedParamPath(
            'enable_user_identification', vartype='yesno',
            path='enable-user-identification'))
        params.append(VersionedParamPath(
            'include_acl', vartype='member',
            path='user-acl/include-list'))
        params.append(VersionedParamPath(
            'exclude_acl', vartype='member',
            path='user-acl/exclude-list'))

        self._params = tuple(params)


class StaticMac(VersionedPanObject):
    """Static MAC address for a Vlan

    Can be added to a :class:`pandevice.network.Vlan` object

    Args:
        interface (str): Name of an interface

    """
    SUFFIX = ENTRY
    NAME = 'mac'

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/mac')

        # params
        params = []

        params.append(VersionedParamPath(
            'interface', path='interface'))

        self._params = tuple(params)


class Vlan(VsysOperations):
    """Vlan

    Args:
        interface (list): List of interface names
        virtual-interface (VlanInterface): The layer3 vlan interface for this vlan

    """
    SUFFIX = ENTRY
    ROOT = Root.DEVICE
    CHILDTYPES = (
        'network.StaticMac',
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/vlan')

        # xpath_imports
        self._xpath_imports.add_profile(value='/network/vlan')

        # params
        params = []

        params.append(VersionedParamPath(
            'interface', vartype='member', path='interface'))
        params.append(VersionedParamPath(
            'virtual_interface', path='/virtual-interface/interface'))

        self._params = tuple(params)


class IPv6Address(VersionedPanObject):
    """IPv6 Address

    Can be added to any :class:`pandevice.network.Interface` subclass
    that supports IPv6

    Args:
        enabled-on-interface (bool): Enabled IPv6 on the interface this
            object was added to
        prefix (bool): Use interface ID as host portion
        anycast (bool): Enable anycast
        advertise_enabled (bool): Enabled router advertisements
        valid_lifetime (int): Valid lifetime
        onlink_flag (bool):
        auto_config_flag (bool):

    """
    SUFFIX = ENTRY
    NAME = "address"

    def _setup(self):
        # xpaths
        # Non-mode interface xpaths
        self._xpaths.add_profile(value='/ipv6/address')
        # Mode interface xpaths (mode: layer3)
        self._xpaths.add_profile(
            value='/layer3/ipv6/address',
            parents=('EthernetInterface', 'AggregateInterface'))

        # params
        params = []

        params.append(VersionedParamPath(
            'enable_on_interface', vartype='yesno',
            path='enable-on-interface'))
        params.append(VersionedParamPath(
            'prefix', vartype='exist', path='prefix'))
        params.append(VersionedParamPath(
            'anycast', vartype='exist', path='anycast'))
        params.append(VersionedParamPath(
            'advertise_enabled', vartype='yesno',
            path='advertise/enable'))
        params.append(VersionedParamPath(
            'valid_lifetime', vartype='int',
            path='advertise/valid-lifetime'))
        params.append(VersionedParamPath(
            'preferred_lifetime', vartype='int',
            path='advertise/preferred-lifetime'))
        params.append(VersionedParamPath(
            'onlink_flag', vartype='yesno',
            path='advertise/onlink-flag'))
        params.append(VersionedParamPath(
            'auto_config_flag', vartype='yesno',
            path='advertise/auto-config-flag'))

        self._params = tuple(params)


class Interface(VsysOperations):
    """Base class for all interfaces

    Do not instantiate this object. Use a subclass.
    Methods in this class are available to all interface subclasses.

    Args:
        name (str): Name of the interface
        state (str): Link state, 'up' or 'down'

    """
    SUFFIX = ENTRY
    ROOT = Root.DEVICE
    DEFAULT_MODE = None
    ALLOW_SET_VLAN = False
    ALWAYS_IMPORT = True

    def up(self):
        """Link state of interface

        Returns:
            bool: True if state is 'up', False if state is 'down',
                'unconfigured' or other

        """
        return self.state == 'up'

    def set_zone(self, zone_name, mode=None, refresh=False,
                 update=False, running_config=False):
        """Set the zone for this interface

        Creates a reference to this interface in the specified zone and removes
        references to this interface from all other zones. The zone will be
        created if it doesn't exist.

        Args:
            zone_name (str): The name of the Zone or a
                :class:`pandevice.network.Zone` instance
            mode (str): The mode of the zone. See
                :class:`pandevice.network.Zone` for possible values
            refresh (bool): Refresh the relevant current state of the device
                before taking action (Default: False)
            update (bool): Apply the changes to the device (Default: False)
            running_config: If refresh is True, refresh from the running
                configuration (Default: False)

        Returns:
            Zone: The zone for this interface after the operation completes

        """
        if mode is None:
            mode = self.DEFAULT_MODE

        return self._set_reference(
            zone_name, Zone, "interface", True, refresh,
            update, running_config, mode=mode)

    def set_virtual_router(self, virtual_router_name, refresh=False, update=False, running_config=False):
        """Set the virtual router for this interface

        Creates a reference to this interface in the specified virtual router
        and removes references to this interface from all other virtual
        routers. The virtual router will be created if it doesn't exist.

        Args:
            virtual_router_name (str): The name of the VirtualRouter or
                a :class:`pandevice.network.VirtualRouter` instance
            refresh (bool): Refresh the relevant current state of the device
                before taking action (Default: False)
            update (bool): Apply the changes to the device (Default: False)
            running_config: If refresh is True, refresh from the running
                configuration (Default: False)

        Returns:
            Zone: The zone for this interface after the operation completes

        """
        return self._set_reference(virtual_router_name, VirtualRouter,
                                   "interface", True, refresh, update,
                                   running_config)

    def set_vlan(self, vlan_name, refresh=False,
                 update=False, running_config=False):
        """Set the vlan for this interface

        Creates a reference to this interface in the specified vlan and removes
        references to this interface from all other interfaces.  The vlan will
        be created if it doesn't exist.

        Args:
            vlan_name (str): The name of the vlan or
                a :class:`pandevice.network.Vlan` instance
            refresh (bool): Refresh the relevant current state of the device
                before taking action (Default: False)
            update (bool): Apply the changes to the device (Default: False)
            running_config: If refresh is True, refresh from the running
                configuration (Default: False)

        Raises:
            AttributeError: if this class is not allowed to use this function.

        Returns:
            Zone: The zone for this interface after the operation completes

        """
        if not self.ALLOW_SET_VLAN:
            msg = 'Class "{0}" cannot invoke this function'
            raise AttributeError(msg.format(self.__class__))

        return self._set_reference(vlan_name, Vlan, "interface", True,
                                   refresh, update, running_config)

    def get_counters(self):
        """Pull the counters for an interface

        Returns:
            dict: counter name as key, counter as value, None if interface is
                not configured

        """
        from pan.config import PanConfig

        device = self.nearest_pandevice()
        cmd = 'show counter interface "{0}"'.format(self.name)
        pconf = device.op(cmd)
        pconf = PanConfig(pconf)
        response = pconf.python()
        logger.debug("response: " + str(response))
        counters = response['response']['result']
        if counters is not None:
            entry = {}
            # Check for entry in ifnet
            if 'entry' in counters.get('ifnet', {}):
                entry = counters['ifnet']['entry'][0]
            elif 'ifnet' in counters.get('ifnet', {}):
                if 'entry' in counters['ifnet'].get('ifnet', {}):
                    entry = counters['ifnet']['ifnet']['entry'][0]

            # Convert strings to integers, if they are integers
            entry.update((k, pandevice.convert_if_int(v))
                         for k, v in entry.items())

            # If empty dictionary (no results) it usually means the interface is not
            # configured, so return None
            return entry if entry else None

    def refresh_state(self):
        """Pull the state of the interface from the firewall

        The attribute 'state' is populated with the current state from the
        firewall.

        Returns:
            str: The current state from the firewall

        """
        device = self.nearest_pandevice()
        cmd = 'show interface "{0}"'.format(self.name)
        response = device.op(cmd)
        state = response.findtext("result/hw/state")
        if state is None:
            state = "unconfigured"
        self.state = state

        return self.state

    def full_delete(self, refresh=False, delete_referencing_objects=False,
                    include_vsys=False):
        """Delete the interface and all references to the interface

        Often when deleting an interface there is an API error because
        there are still references to the interface from zones, virtual-router,
        vsys, etc. This method deletes all references to the interface before
        deleting the interface itself.

        Args:
            refresh (bool): Refresh the current state of the device before
                taking action
            delete_referencing_objects (bool): Delete the entire object that
                references this interface

        """
        self.set_zone(None, refresh=refresh, update=True)
        if self.ALLOW_SET_VLAN:
            self.set_vlan(None, refresh=refresh, update=True)
        self.set_virtual_router(None, refresh=refresh, update=True)

        # Remove any references to the interface across all known
        # children of this pan_device. This does not use 'refresh'.
        # Only pre-refreshed objects are scanned for references.
        for obj in self.nearest_pandevice().findall(PanObject, recursive=True):
            if isinstance(obj, device.Vsys):
                if not include_vsys:
                    continue
            try:
                if str(self) == obj.interface or self == obj.interface:
                    if delete_referencing_objects:
                        obj.delete()
                    else:
                        obj.interface = None
                        obj.update("interface")
                elif "__iter__" in dir(obj.interface) and str(self) in obj.interface:
                    if delete_referencing_objects:
                        obj.delete()
                    else:
                        obj.interface.remove(str(self))
                        obj.update("interface")
                elif "__iter__" in dir(obj.interface) and self in obj.interface:
                    if delete_referencing_objects:
                        obj.delete()
                    else:
                        obj.interface.remove(self)
                        obj.update("interface")
            except AttributeError:
                pass
        self.delete()


class Arp(VersionedPanObject):
    """Static ARP Mapping

    Can be added to various interfaces.

    Args:
        ip (str): The IP address
        hw_address (str): The MAC address for the static ARP
        interface (str): The interface (when attached to VlanInterface only)

    """
    SUFFIX = ENTRY
    NAME = 'ip'

    def _setup(self):
        # xpaths
        # Interface xpaths
        self._xpaths.add_profile(value='/layer3/arp')
        # Subinterface xpaths
        self._xpaths.add_profile(
            value='/arp',
            parents=('Layer3Subinterface', ))

        # params
        params = []

        params.append(VersionedParamPath(
            'hw_address', path='hw-address'))
        params.append(VersionedParamPath(
            'interface', path='interface'))

        self._params = tuple(params)


class VirtualWire(VersionedPanObject):
    """Virtual wires (vwire)

    Args:
        name (str): The vwire name
        tag (int): Tag for the interface, aka vlan id
        interface1 (str): The first interface to use
        interface2 (str): The second interface to use
        multicast (bool): Enable multicast firewalling or not
        pass_through (bool): Enable link state pass through or not

    """
    ROOT = Root.DEVICE
    SUFFIX = ENTRY

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/virtual-wire')

        # params
        params = []

        params.append(VersionedParamPath(
            'tag', path='tag-allowed', vartype='int'))
        params.append(VersionedParamPath(
            'interface1', path='interface1'))
        params.append(VersionedParamPath(
            'interface2', path='interface2'))
        params.append(VersionedParamPath(
            'multicast', path='multicast-firewalling/enable',
            default=False, vartype='yesno'))
        params.append(VersionedParamPath(
            'pass_through', path='link-state-pass-through/enable',
            default=True, vartype='yesno'))

        self._params = tuple(params)


class Subinterface(Interface):
    """Subinterface class

    Do not instantiate this object. Use a subclass.

    """
    _BASE_INTERFACE_NAME = 'entry BASE_INTERFACE_NAME'

    def set_name(self):
        """Create a name appropriate for a subinterface if it isn't already"""
        if '.' not in self.name:
            self.name = '{0}.{1}'.format(self.name, self.tag)

    @property
    def XPATH(self):
        path = super(Subinterface, self).XPATH

        if self._BASE_INTERFACE_NAME in path:
            base = self.uid.split('.')[0]
            path = path.replace(self._BASE_INTERFACE_NAME,
                                "entry[@name='{0}']".format(base))

        return path


class AbstractSubinterface(object):
    """When a subinterface is needed, but the layer is unknown

    Kindof like a placeholder or reference for a Layer2Subinterface or Layer3Subinterface.
    This class gets a parent which is the ethernet or aggregate interface, but it should
    not be added to the parent interface with add().

    Args:
        name (str): Name of the interface (eg. ethernet1/1.5)
        tag (int): Tag for the interface, aka vlan id
        parent (Interface): The base interface for this subinterface

    """
    def __init__(self, name, tag, parent=None):
        self.name = name
        self.tag = tag
        self.parent = parent

    def set_name(self):
        """Create a name appropriate for a subinterface if it isn't already created

        Example:
            If self.name is 'ethernet1/1' and self.tag is 5, this method will change the
            name to 'ethernet1/1.5'.

        """
        if self.name.find(".") == -1:
            self.name = self.name + "." + str(self.tag)

    def nearest_pandevice(self):
        """The PanDevice parent for this instance

        Returns:
            PanDevice: Parent PanDevice instance (Firewall or Panorama)

        """
        return self.parent._nearest_pandevice()

    def set_zone(self, zone_name, mode=None, refresh=False, update=False, running_config=False):
        raise err.PanDeviceError("Unable to set zone on abstract subinterface because layer must be known to set zone")

    def set_virtual_router(self, virtual_router_name, refresh=False, update=False, running_config=False):
        """Set the virtual router for this interface

        Creates a reference to this interface in the specified virtual router and removes references
        to this interface from all other virtual routers. The virtual router will be created if it doesn't exist.

        Args:
            virtual_router_name (str): The name of the VirtualRouter or
                a :class:`pandevice.network.VirtualRouter` instance
            refresh (bool): Refresh the relevant current state of the device before taking action
                (Default: False)
            update (bool): Apply the changes to the device (Default: False)
            running_config: If refresh is True, refresh from the running configuration
                (Default: False)

        Returns:
            Zone: The zone for this interface after the operation completes

        """
        interface = Layer3Subinterface(self.name, self.tag)
        interface.parent = self.parent
        return interface._set_reference(virtual_router_name, VirtualRouter, "interface", True, refresh=False, update=update, running_config=running_config)

    def get_layered_subinterface(self, mode, add=True):
        """Instantiate a regular subinterface type from this AbstractSubinterface

        Converts an abstract subinterface to a real subinterface by offering it a mode.

        Args:
            mode (str): Mode of the subinterface ('layer3' or 'layer2')
            add (bool): Add the newly instantiated subinterface to the base interface object

        Returns:
            Subinterface: A :class:`pandevice.network.Layer3Subinterface` or
            :class:`pandevice.network.Layer2Subinterface` instance, depending on the mode argument

        """
        if self.parent is not None:
            if mode == "layer3":
                subintclass = Layer3Subinterface
            elif mode == "layer2":
                subintclass = Layer2Subinterface
            else:
                raise err.PanDeviceError("Unknown layer passed to subinterface factory: %s" % mode)
            layered_subinterface = self.parent.find(self.name, subintclass)
            # Verify tag is correct
            if layered_subinterface is not None:
                if layered_subinterface.tag != self.tag:
                    layered_subinterface.tag = self.tag
            else:
                if add:
                    layered_subinterface = self.parent.add(subintclass(self.name, tag=self.tag))
                else:
                    return
            return layered_subinterface

    def delete(self):
        """Deletes both Layer3 and Layer2 subinterfaces by name

        This is necessary because an AbstractSubinterface's mode is unknown.

        """
        for cls in (Layer3Subinterface, Layer2Subinterface):
            i = self.parent.find_or_create(self.uid, cls, tag=self.tag)
            i.delete()


class Layer3Subinterface(Subinterface):
    """Ethernet or Aggregate Subinterface in Layer 3 mode.

    Args:
        tag (int): Tag for the interface, aka vlan id
        ip (tuple): Interface IPv4 addresses
        ipv6_enabled (bool): IPv6 Enabled (requires IPv6Address child object)
        management_profile (ManagementProfile): Interface Management Profile
        mtu(int): MTU for interface
        adjust_tcp_mss (bool): Adjust TCP MSS
        netflow_profile (NetflowProfile): Netflow profile
        comment (str): The interface's comment
        ipv4_mss_adjust(int): TCP MSS adjustment for ipv4
        ipv6_mss_adjust(int): TCP MSS adjustment for ipv6
        enable_dhcp (bool): Enable DHCP on this interface
        create_dhcp_default_route (bool): Create default route pointing to default gateway provided by server
        dhcp_default_route_metric (int): Metric for the DHCP default route

    """
    DEFAULT_MODE = 'layer3'
    CHILDTYPES = (
        "network.IPv6Address",
        "network.Arp",
    )

    def _setup(self):
        # xpaths for parents: EthernetInterface, AggregateInterface)
        self._xpaths.add_profile(value='/layer3/units')
        # xpaths for parents: firewall.Firewall, device.Vsys
        self._xpaths.add_profile(
            parents=('Firewall', 'Vsys'),
            value=('/network/interface/ethernet/{0}/layer3/units'.format(
                self._BASE_INTERFACE_NAME)))

        # xpath imports
        self._xpath_imports.add_profile(value='/network/interface')

        # params
        params = []

        params.append(VersionedParamPath(
            'tag', path='tag', vartype='int'))
        params.append(VersionedParamPath(
            'ip', path='ip', vartype='entry'))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='ipv6/enabled', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno',
            path='ipv6/neighbor-discovery/router-advertisement/enable')
        params.append(VersionedParamPath(
            'management_profile', path='interface-management-profile'))
        params.append(VersionedParamPath(
            'mtu', path='mtu', vartype='int'))
        params.append(VersionedParamPath(
            'adjust_tcp_mss', path='adjust-tcp-mss', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno', path='adjust-tcp-mss/enable')
        params.append(VersionedParamPath(
            'netflow_profile', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))
        params.append(VersionedParamPath(
            'ipv4_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv4-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'ipv6_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv6-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'enable_dhcp', path='dhcp-client/enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'create_dhcp_default_route',
            path='dhcp-client/create-default-route', vartype='yesno'))
        params.append(VersionedParamPath(
            'dhcp_default_route_metric',
            path='dhcp-client/default-route-metric', vartype='int'))

        self._params = tuple(params)


class Layer2Subinterface(Subinterface):
    """Ethernet or Aggregate Subinterface in Layer 2 mode.

    Args:
        tag (int): Tag for the interface, aka vlan id
        lldp_enabled (bool): Enable LLDP
        lldp_profile (str): Reference to an lldp profile
        netflow_profile_l2 (NetflowProfile): Reference to a netflow profile
        comment (str): The interface's comment

    """
    SUFFIX = ENTRY
    DEFAULT_MODE = 'layer2'
    ALLOW_SET_VLAN = True

    def _setup(self):
        # xpaths for parents: EthernetInterface, AggregateInterface
        self._xpaths.add_profile(value='/layer2/units')
        # xpaths for parents: firewall.Firewall, device.Vsys
        self._xpaths.add_profile(
            parents=('Firewall', 'Vsys'),
            value=('/network/interface/ethernet/{0}/layer2/units'.format(
                self._BASE_INTERFACE_NAME)))

        # xpath imports
        self._xpath_imports.add_profile(value='/network/interface')

        # params
        params = []

        params.append(VersionedParamPath(
            'tag', path='tag', vartype='int'))
        params.append(VersionedParamPath(
            'lldp_enabled', path='lldp/enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'lldp_profile', path='lldp/profile'))
        params.append(VersionedParamPath(
            'netflow_profile_l2', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))

        self._params = tuple(params)


class PhysicalInterface(Interface):
    """Absract base class for Ethernet and Aggregate Interfaces

    Do not instantiate this object. Use a subclass.

    """
    def set_zone(self, zone_name, mode=None, refresh=False,
                 update=False, running_config=False):
        """Set the zone for this interface

        Creates a reference to this interface in the specified zone and removes
        references to this interface from all other zones. The zone will be
        created if it doesn't exist.

        Args:
            zone_name (str): The name of the Zone or a
                :class:`pandevice.network.Zone` instance
            mode (str): The mode of the zone. See
                :class:`pandevice.network.Zone` for possible values
            refresh (bool): Refresh the relevant current state of the device
                before taking action (Default: False)
            update (bool): Apply the changes to the device (Default: False)
            running_config: If refresh is True, refresh from the running
                configuration (Default: False)

        Returns:
            Zone: The zone for this interface after the operation completes

        """
        if mode is None:
            mode = self.mode

        return super(PhysicalInterface, self).set_zone(
            zone_name, mode, refresh, update, running_config)


class EthernetInterface(PhysicalInterface):
    """Ethernet interface (eg. 'ethernet1/1')

    Args:
        name (str): Name of interface (eg. 'ethernet1/1')
        mode (str): Mode of the interface:
                * layer3
                * layer2
                * virtual-wire
                * tap
                * ha
                * decrypt-mirror
                * aggregate-group

            Not all modes apply to all interface types (Default: layer3)

        ip (tuple): Layer3: Interface IPv4 addresses
        ipv6_enabled (bool): Layer3: IPv6 Enabled (requires
            IPv6Address child object)
        management_profile (ManagementProfile): Layer3: Interface Management
            Profile
        mtu(int): Layer3: MTU for interface
        adjust_tcp_mss (bool): Layer3: Adjust TCP MSS
        netflow_profile (NetflowProfile): Netflow profile
        lldp_enabled (bool): Layer2: Enable LLDP
        lldp_profile (str): Layer2: Reference to an lldp profile
        netflow_profile_l2 (NetflowProfile): Netflow profile
        link_speed (str): Link speed: eg. auto, 10, 100, 1000
        link_duplex (str): Link duplex: eg. auto, full, half
        link_state (str): Link state: eg. auto, up, down
        aggregate_group (str): Aggregate interface (eg. ae1)
        comment (str): The interface's comment
        ipv4_mss_adjust(int): TCP MSS adjustment for ipv4
        ipv6_mss_adjust(int): TCP MSS adjustment for ipv6
        enable_dhcp (bool): Enable DHCP on this interface
        create_dhcp_default_route (bool): Create default route pointing to default gateway provided by server
        dhcp_default_route_metric (int): Metric for the DHCP default route

    """
    ALLOW_SET_VLAN = True
    CHILDTYPES = (
        "network.Layer3Subinterface",
        "network.Layer2Subinterface",
        "network.IPv6Address",
        "network.Arp",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/interface/ethernet')

        # xpath imports
        self._xpath_imports.add_profile(value='/network/interface')

        # params
        params = []

        params.append(VersionedParamPath(
            'mode', path='{mode}', default='layer3',
            values=[
                'layer3', 'layer2', 'virtual-wire', 'tap',
                'ha', 'decrypt-mirror', 'aggregate-group',
            ]))
        params.append(VersionedParamPath(
            'ip', path='{mode}/ip', vartype='entry',
            condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='{mode}/ipv6/enabled', vartype='yesno',
            condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'management_profile', path='{mode}/interface-management-profile',
            condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'mtu', path='{mode}/mtu', vartype='int',
            condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'adjust_tcp_mss', path='{mode}/adjust-tcp-mss', vartype='yesno',
            condition={'mode': 'layer3'}))
        params[-1].add_profile(
            '7.1.0',
            path='{mode}/adjust-tcp-mss/enable',
            vartype='yesno', condition={'mode': 'layer3'})
        params.append(VersionedParamPath(
            'netflow_profile', path='{mode}/netflow-profile',
            condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'lldp_enabled', path='{mode}/lldp/enable', vartype='yesno',
            condition={'mode': 'layer2'}))
        params.append(VersionedParamPath(
            'lldp_profile', path='{mode}/lldp/profile',
            condition={'mode': 'layer2'}))
        params.append(VersionedParamPath(
            'netflow_profile_l2', path='{mode}/netflow-profile',
            condition={'mode': 'layer2'}))
        params.append(VersionedParamPath(
            'link_speed', path='link-speed'))
        params.append(VersionedParamPath(
            'link_duplex', path='link-duplex'))
        params.append(VersionedParamPath(
            'link_state', path='link-state'))
        params.append(VersionedParamPath(
            'aggregate_group', path='aggregate-group',
            condition={'mode': 'aggregate-group'}))
        params.append(VersionedParamPath(
            'comment', path='comment'))
        params.append(VersionedParamPath(
            'ipv4_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='{mode}/adjust-tcp-mss/ipv4-mss-adjustment',
            vartype='int', condition={'mode': 'layer3'})
        params.append(VersionedParamPath(
            'ipv6_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='{mode}/adjust-tcp-mss/ipv6-mss-adjustment',
            vartype='int', condition={'mode': 'layer3'})
        params.append(VersionedParamPath(
            'enable_dhcp', path='{mode}/dhcp-client/enable',
            vartype='yesno', condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'create_dhcp_default_route',
            path='{mode}/dhcp-client/create-default-route',
            vartype='yesno', condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'dhcp_default_route_metric',
            path='{mode}/dhcp-client/default-route-metric',
            vartype='int', condition={'mode': 'layer3'}))

        self._params = tuple(params)


class AggregateInterface(PhysicalInterface):
    """Aggregate interface (eg. 'ae1')

    Args:
        name (str): Name of interface (eg. 'ae1')
        mode (str): Mode of the interface:
                * layer3
                * layer2
                * virtual-wire
                * tap
                * ha
                * decrypt-mirror
                * aggregate-group

            Not all modes apply to all interface types (Default: layer3)

        ip (tuple): Layer3: Interface IPv4 addresses
        ipv6_enabled (bool): Layer3: IPv6 Enabled (requires
            IPv6Address child object)
        management_profile (ManagementProfile): Layer3: Interface Management
            Profile
        mtu(int): Layer3: MTU for interface
        adjust_tcp_mss (bool): Layer3: Adjust TCP MSS
        netflow_profile (NetflowProfile): Netflow profile
        lldp_enabled (bool): Layer2: Enable LLDP
        lldp_profile (str): Layer2: Reference to an lldp profile
        netflow_profile_l2 (NetflowProfile): Netflow profile
        comment (str): The interface's comment
        ipv4_mss_adjust(int): TCP MSS adjustment for ipv4
        ipv6_mss_adjust(int): TCP MSS adjustment for ipv6
        enable_dhcp (bool): Enable DHCP on this interface
        create_dhcp_default_route (bool): Create default route pointing to default gateway provided by server
        dhcp_default_route_metric (int): Metric for the DHCP default route

    """
    ALLOW_SET_VLAN = True
    CHILDTYPES = (
        "network.Layer3Subinterface",
        "network.Layer2Subinterface",
        "network.IPv6Address",
        "network.Arp",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/interface/aggregate-ethernet')

        # xpath imports
        self._xpath_imports.add_profile(value='/network/interface')

        # params
        params = []

        params.append(VersionedParamPath(
            'mode', path='{mode}', default='layer3',
            values=[
                'layer3', 'layer2', 'virtual-wire', 'tap',
                'ha', 'decrypt-mirror', 'aggregate-group',
            ]))
        params.append(VersionedParamPath(
            'ip', path='ip', vartype='entry'))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='ipv6/enabled', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno',
            path='ipv6/neighbor-discovery/router-advertisement/enable')
        params.append(VersionedParamPath(
            'management_profile', path='interface-management-profile'))
        params.append(VersionedParamPath(
            'mtu', path='mtu', vartype='int'))
        params.append(VersionedParamPath(
            'adjust_tcp_mss', path='adjust-tcp-mss', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno', path='adjust-tcp-mss/enable')
        params.append(VersionedParamPath(
            'netflow_profile', path='netflow-profile'))
        params.append(VersionedParamPath(
            'lldp_enabled', path='lldp/enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'lldp_profile', path='lldp/profile'))
        params.append(VersionedParamPath(
            'netflow_profile_l2', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))
        params.append(VersionedParamPath(
            'ipv4_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv4-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'ipv6_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv6-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'enable_dhcp', path='{mode}/dhcp-client/enable',
            vartype='yesno', condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'create_dhcp_default_route',
            path='{mode}/dhcp-client/create-default-route',
            vartype='yesno', condition={'mode': 'layer3'}))
        params.append(VersionedParamPath(
            'dhcp_default_route_metric',
            path='{mode}/dhcp-client/default-route-metric',
            vartype='int', condition={'mode': 'layer3'}))

        self._params = tuple(params)


class VlanInterface(Interface):
    """Vlan interface

    Args:
        ip (tuple): Interface IPv4 addresses
        ipv6_enabled (bool): IPv6 Enabled (requires IPv6Address child object)
        management_profile (ManagementProfile): Interface Management Profile
        mtu(int): MTU for interface
        adjust_tcp_mss (bool): Adjust TCP MSS
        netflow_profile (NetflowProfile): Netflow profile
        comment (str): The interface's comment
        ipv4_mss_adjust(int): TCP MSS adjustment for ipv4
        ipv6_mss_adjust(int): TCP MSS adjustment for ipv6
        enable_dhcp (bool): Enable DHCP on this interface
        create_dhcp_default_route (bool): Create default route pointing to default gateway provided by server
        dhcp_default_route_metric (int): Metric for the DHCP default route

    """
    CHILDTYPES = (
        "network.IPv6Address",
        "network.Arp",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/interface/vlan/units')

        # params
        params = []

        params.append(VersionedParamPath(
            'ip', path='ip', vartype='entry'))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='ipv6/enabled', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno',
            path='ipv6/neighbor-discovery/router-advertisement/enable')
        params.append(VersionedParamPath(
            'management_profile', path='interface-management-profile'))
        params.append(VersionedParamPath(
            'mtu', path='mtu', vartype='int'))
        params.append(VersionedParamPath(
            'adjust_tcp_mss', path='adjust-tcp-mss', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno', path='adjust-tcp-mss/enable')
        params.append(VersionedParamPath(
            'netflow_profile', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))
        params.append(VersionedParamPath(
            'ipv4_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv4-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'ipv6_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv6-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'enable_dhcp', path='dhcp-client/enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'create_dhcp_default_route',
            path='dhcp-client/create-default-route', vartype='yesno'))
        params.append(VersionedParamPath(
            'dhcp_default_route_metric',
            path='dhcp-client/default-route-metric', vartype='int'))

        self._params = tuple(params)


class LoopbackInterface(Interface):
    """Loopback interface

    Args:
        ip (tuple): Interface IPv4 addresses
        ipv6_enabled (bool): IPv6 Enabled (requires IPv6Address child object)
        management_profile (ManagementProfile): Interface Management Profile
        mtu(int): MTU for interface
        adjust_tcp_mss (bool): Adjust TCP MSS
        netflow_profile (NetflowProfile): Netflow profile
        comment (str): The interface's comment
        ipv4_mss_adjust(int): TCP MSS adjustment for ipv4
        ipv6_mss_adjust(int): TCP MSS adjustment for ipv6

    """
    CHILDTYPES = (
        "network.IPv6Address",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/interface/loopback/units')

        # params
        params = []

        params.append(VersionedParamPath(
            'ip', path='ip', vartype='entry'))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='ipv6/enabled', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno',
            path='ipv6/neighbor-discovery/router-advertisement/enable')
        params.append(VersionedParamPath(
            'management_profile', path='interface-management-profile'))
        params.append(VersionedParamPath(
            'mtu', path='mtu', vartype='int'))
        params.append(VersionedParamPath(
            'adjust_tcp_mss', path='adjust-tcp-mss', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno', path='adjust-tcp-mss/enable')
        params.append(VersionedParamPath(
            'netflow_profile', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))
        params.append(VersionedParamPath(
            'ipv4_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv4-mss-adjustment', vartype='int')
        params.append(VersionedParamPath(
            'ipv6_mss_adjust', exclude=True))
        params[-1].add_profile(
            '7.1.0',
            path='adjust-tcp-mss/ipv6-mss-adjustment', vartype='int')

        self._params = tuple(params)


class TunnelInterface(Interface):
    """Tunnel interface

    Args:
        ip (tuple): Interface IPv4 addresses
        ipv6_enabled (bool): IPv6 Enabled (requires IPv6Address child object)
        management_profile (ManagementProfile): Interface Management Profile
        mtu(int): MTU for interface
        netflow_profile (NetflowProfile): Netflow profile
        comment (str): The interface's comment

    """
    CHILDTYPES = (
        "network.IPv6Address",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/interface/tunnel/units')

        # params
        params = []

        params.append(VersionedParamPath(
            'ip', path='ip', vartype='entry'))
        params.append(VersionedParamPath(
            'ipv6_enabled', path='ipv6/enabled', vartype='yesno'))
        params[-1].add_profile(
            '7.1.0',
            vartype='yesno',
            path='ipv6/neighbor-discovery/router-advertisement/enable')
        params.append(VersionedParamPath(
            'management_profile', path='interface-management-profile'))
        params.append(VersionedParamPath(
            'mtu', path='mtu', vartype='int'))
        params.append(VersionedParamPath(
            'netflow_profile', path='netflow-profile'))
        params.append(VersionedParamPath(
            'comment', path='comment'))

        self._params = tuple(params)


class StaticRoute(VersionedPanObject):
    """Static Route

    Add to a :class:`pandevice.network.VirtualRouter` instance.

    Args:
        name (str): The name
        destination (str): Destination network
        nexthop_type (str): ip-address or discard
        nexthop (str): Next hop IP address
        interface (str): Next hop interface
        admin_dist (str): Administrative distance
        metric (int): Metric (Default: 10)

    """
    SUFFIX = ENTRY

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/routing-table/ip/static-route')

        # params
        params = []

        params.append(VersionedParamPath(
            'destination', path='destination'))
        params.append(VersionedParamPath(
            'nexthop_type', default='ip-address',
            values=['discard', 'ip-address'],
            path='nexthop/{nexthop_type}'))
        params.append(VersionedParamPath(
            'nexthop', path='nexthop/{nexthop_type}'))
        params.append(VersionedParamPath(
            'interface', path='interface'))
        params.append(VersionedParamPath(
            'admin_dist', vartype='int', path='admin-dist'))
        params.append(VersionedParamPath(
            'metric', default=10, vartype='int', path='metric'))

        self._params = tuple(params)


class StaticRouteV6(VersionedPanObject):
    """IPV6 Static Route

    Add to a :class:`pandevice.network.VirtualRouter` instance.

    Args:
        name (str): The name
        destination (str): Destination network
        nexthop_type (str): ip-address or discard
        nexthop (str): Next hop IP address
        interface (str): Next hop interface
        admin_dist (str): Administrative distance
        metric (int): Metric (Default: 10)

    """
    SUFFIX = ENTRY

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/routing-table/ipv6/static-route')

        # params
        params = []

        params.append(VersionedParamPath(
            'destination', path='destination'))
        params.append(VersionedParamPath(
            'nexthop_type', default='ipv6-address',
            values=['discard', 'ipv6-address'],
            path='nexthop/{nexthop_type}'))
        params.append(VersionedParamPath(
            'nexthop', path='nexthop/{nexthop_type}'))
        params.append(VersionedParamPath(
            'interface', path='interface'))
        params.append(VersionedParamPath(
            'admin_dist', vartype='int', path='admin-dist'))
        params.append(VersionedParamPath(
            'metric', default=10, vartype='int', path='metric'))

        self._params = tuple(params)

class VirtualRouter(VsysOperations):
    """Virtual router

    Args:
        name (str): Name of virtual router (Default: "default")
        interface (list): List of interface names
        ad_static (int): Administrative distance for this protocol
        ad_static_ipv6 (int): Administrative distance for this protocol
        ad_ospf_int (int): Administrative distance for this protocol
        ad_ospf_ext (int): Administrative distance for this protocol
        ad_ospfv3_int (int): Administrative distance for this protocol
        ad_ospfv3_ext (int): Administrative distance for this protocol
        ad_ibgp (int): Administrative distance for this protocol
        ad_ebgp (int): Administrative distance for this protocol
        ad_rip (int): Administrative distance for this protocol

    """
    _DEFAULT_NAME = 'default'
    SUFFIX = ENTRY
    CHILDTYPES = (
        "network.StaticRoute",
        "network.StaticRouteV6",
        "network.RedistributionProfile",
        "network.Ospf",
    )

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(value='/network/virtual-router')

        # xpath imports
        self._xpath_imports.add_profile(value='/network/virtual-router')

        # params
        params = []

        params.append(VersionedParamPath(
            'interface', path='interface', vartype='member'))

        admin_dists = (
            ('ad_static', 'static'), ('ad_static_ipv6', 'static-ipv6'),
            ('ad_ospf_int', 'ospf-int'), ('ad_ospf_ext', 'ospf-ext'),
            ('ad_ospfv3_int', 'ospfv3-int'), ('ad_ospfv3_ext', 'ospfv3-ext'),
            ('ad_ibgp', 'ibgp'), ('ad_ebgp', 'ebgp'), ('ad_rip', 'rip')
        )

        for var_name, path in admin_dists:
            params.append(VersionedParamPath(
                var_name, vartype='int', path='admin-dists/' + path))

        self._params = tuple(params)


class RedistributionProfile(VersionedPanObject):
    """Redistribution Profile

    Args:
        name (str): Name of profile
        priority (int): Priority id
        action (str): 'no-redist' or 'redist'
        filter_type (tuple): Any of 'static', 'connect', 'rip', 'ospf', or 'bgp'
        filter_interface (tuple): Filter interface
        filter_destination (tuple): Filter destination
        filter_nexthop (tuple): Filter nexthop
        ospf_filter_pathtype (tuple): Any of 'intra-area', 'inter-area', 'ext-1', or 'ext-2
        ospf_filter_area (tuple): OSPF filter on area
        ospf_filter_tag (tuple): OSPF filter on tag
        bgp_filter_community (tuple): BGP filter on community
        bgp_filter_extended_community (tuple): BGP filter on extended community

    """
    SUFFIX = ENTRY

    def _setup(self):
        self._xpaths.add_profile(value='/protocol/redist-profile')

        params = []

        params.append(VersionedParamPath(
            'priority', vartype='int'))
        params.append(VersionedParamPath(
            'action', values=['no-redist', 'redist'], path='action/{action}'))
        params.append(VersionedParamPath(
            'filter_type', path='filter/type', vartype='member'))
        params.append(VersionedParamPath(
            'filter_interface', path='filter/interface', vartype='member'))
        params.append(VersionedParamPath(
            'filter_destination', path='filter/destination', vartype='member'))
        params.append(VersionedParamPath(
            'filter_nexthop', path='filter/nexthop', vartype='member'))
        params.append(VersionedParamPath(
            'ospf_filter_pathtype', path='filter/ospf/path-type',
            vartype='member'))
        params.append(VersionedParamPath(
            'ospf_filter_area', path='filter/ospf/area', vartype='member'))
        params.append(VersionedParamPath(
            'ospf_filter_tag', path='filter/ospf/tag', vartype='member'))
        params.append(VersionedParamPath(
            'bgp_filter_community', path='filter/bgp/community',
            vartype='member'))
        params.append(VersionedParamPath(
            'bgp_filter_extended_community',
            path='filter/bgp/extended-community', vartype='member'))

        self._params = tuple(params)


class Ospf(VersionedPanObject):
    """OSPF Process

    Args:
        enable (bool): Enable OSPF (Default: True)
        router_id (str): Router ID in IP format (eg. 1.1.1.1)
        reject_default_route (bool): Reject default route
        allow_redist_default_route (bool): Allow redistribution in default route
        rfc1583 (bool): rfc1583
        spf_calculation_delay (int): SPF calculation delay
        lsa_interval (int): LSA interval
        graceful_restart_enable (bool): Enable OSPF graceful restart
        gr_grace_period (int): Graceful restart period
        gr_helper_enable (bool): Graceful restart helper enable
        gr_strict_lsa_checking (bool): Graceful restart strict lsa checking
        gr_max_neighbor_restart_time (int): Graceful restart neighbor restart time

    """
    NAME = None
    CHILDTYPES = (
        "network.OspfArea",
        "network.OspfAuthProfile",
        "network.OspfExportRules",
    )

    def _setup(self):
        self._xpaths.add_profile(value='/protocol/ospf')

        params = []

        params.append(VersionedParamPath(
            'enable', default=True, path='enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'router_id'))
        params.append(VersionedParamPath(
            'reject_default_route', vartype='yesno'))
        params.append(VersionedParamPath(
            'allow_redist_default_route', vartype='yesno'))
        params.append(VersionedParamPath(
            'rfc1583', vartype='yesno'))
        # TODO: Add flood prevention
        params.append(VersionedParamPath(
            'spf_calculation_delay', path='timers/spf-calculation-delay',
            vartype='int'))
        params.append(VersionedParamPath(
            'lsa_interval', path='timers/lsa-interval', vartype='int'))
        params.append(VersionedParamPath(
            'graceful_restart_enable', path='graceful-restart/enable',
            vartype='yesno'))
        params.append(VersionedParamPath(
            'gr_grace_period', path='graceful-restart/grace-period',
            vartype='int'))
        params.append(VersionedParamPath(
            'gr_helper_enable', path='graceful-restart/helper-enable',
            vartype='yesno'))
        params.append(VersionedParamPath(
            'gr_strict_lsa_checking',
            path='graceful-restart/strict-LSA-checking', vartype='yesno'))
        params.append(VersionedParamPath(
            'gr_max_neighbor_restart_time',
            path='graceful-restart/max-neighbor-restart-time', vartype='int'))

        self._params = tuple(params)


class OspfArea(VersionedPanObject):
    """OSPF Area

    Args:
        name (str): Area in IP format
        type (str): Type of area, 'normal', 'stub', or 'nssa' (Default: normal)
        accept_summary (bool): Accept summary route - stub and nssa only
        default_route_advertise (str): 'disable' or 'advertise' (Default: disable) - stub and nssa only
        default_route_advertise_metric (int): Default route metric - stub and nssa only
        default_route_advertise_type (str): 'ext-1' or 'ext2' (Default: ext-2 - nssa only

    """
    SUFFIX = ENTRY
    CHILDTYPES = (
        "network.OspfRange",
        "network.OspfAreaInterface",
        "network.OspfNssaExternalRange",
    )

    def _setup(self):
        self._xpaths.add_profile(value='/area')

        params = []

        params.append(VersionedParamPath(
            'type', default='normal', values=['normal', 'stub', 'nssa'],
            path='type/{type}'))
        params.append(VersionedParamPath(
            'accept_summary',
            condition={'type': ['stub', 'nssa']},
            path='type/{type}/accept-summary',
            vartype='yesno'))
        params.append(VersionedParamPath(
            'default_route_advertise',
            default='disable',
            condition={'type': ['stub', 'nssa']},
            values=['disable', 'advertise'],
            path='type/{type}/default-route/{default_route_advertise}'))
        params.append(VersionedParamPath(
            'default_route_advertise_metric',
            condition={
                'type': ['stub', 'nssa'],
                'default_route_advertise': 'advertise'},
            path='type/{type}/default-route/advertise/metric',
            vartype='int'))
        params.append(VersionedParamPath(
            'default_route_advertise_type',
            default='ext-2',
            condition={'type': 'nssa', 'default_route_advertise': 'advertise'},
            values=['ext-1', 'ext-2'],
            path='type/nssa/default-route/advertise/type'))

        self._params = tuple(params)


class OspfRange(VersionedPanObject):
    """OSPF Range

    Args:
        name (str): IP network with prefix
        mode (str): 'advertise' or 'suppress' (Default: advertise)

    """
    SUFFIX = ENTRY

    def _setup(self):
        self._xpaths.add_profile(value='/range')

        params = []

        params.append(VersionedParamPath(
            'mode', default='advertise', values=['advertise', 'suppress'],
            path='{mode}'))

        self._params = tuple(params)


class OspfNssaExternalRange(VersionedPanObject):
    """OSPF NSSA External Range

    Args:
        name (str): IP network with prefix
        mode (str): 'advertise' or 'suppress' (Default: advertise)

    """
    SUFFIX = ENTRY

    def _setup(self):
        self._xpaths.add_profile(value='/type/nssa/nssa-ext-range')

        params = []

        params.append(VersionedParamPath(
            'mode', default='advertise', values=['advertise', 'suppress'],
            path='{mode}'))

        self._params = tuple(params)


class OspfAreaInterface(VersionedPanObject):
    """OSPF Area Interface

    Args:
        name (str): Name of the interface (interface must exist)
        enable (bool): OSPF enabled on this interface
        passive (bool): Passive mode
        link_type (str): Link type, 'broadcast', 'p2p', or 'p2mp' (Default: broadcast)
        metric (int): Metric
        priority (int): Priority id
        hello_interval (int): Hello interval
        dead_counts (int): Dead counts
        retransmit_interval (int): Retransmit interval
        transit_delay (int): Transit delay
        gr_delay (int): Graceful restart delay
        authentication (str): Reference to a :class:`pandevice.network.OspfAuthProfile`

    """
    SUFFIX = ENTRY
    CHILDTYPES = (
        "network.OspfNeighbor",
    )

    def _setup(self):
        self._xpaths.add_profile(value='/interface')

        params = []

        params.append(VersionedParamPath(
            'enable', vartype='yesno'))
        params.append(VersionedParamPath(
            'passive', vartype='yesno'))
        params.append(VersionedParamPath(
            'link_type', default='broadcast',
            values=['broadcast', 'p2p', 'p2mp'], path='link-type/{link_type}'))
        params.append(VersionedParamPath(
            'metric', vartype='int'))
        params.append(VersionedParamPath(
            'priority', vartype='int'))
        params.append(VersionedParamPath(
            'hello_interval', vartype='int'))
        params.append(VersionedParamPath(
            'dead_counts', vartype='int'))
        params.append(VersionedParamPath(
            'retransmit_interval', vartype='int'))
        params.append(VersionedParamPath(
            'transit_delay', vartype='int'))
        params.append(VersionedParamPath(
            'gr_delay', vartype='int'))
        params.append(VersionedParamPath(
            'authentication'))

        self._params = tuple(params)


class OspfNeighbor(VersionedPanObject):
    """OSPF Neighbor

    Args:
        name (str): IP of neighbor
        metric (int): Metric

    """
    SUFFIX = ENTRY

    def _setup(self):
        self._xpaths.add_profile(value='/neighbor')

        params = []

        params.append(VersionedParamPath(
            'metric', vartype='int', exclude=True))

        self._params = tuple(params)


class OspfAuthProfile(VersionedPanObject):
    """OSPF Authentication Profile

    Args:
        name (str): Name of Auth Profile
        type (str): 'password' or 'md5'
        password (str): The password if type is set to 'password'.
            If type is set to 'md5', add a :class:`pandevice.network.OspfAuthProfileMd5`

    """
    SUFFIX = ENTRY
    CHILDTYPES = (
        "network.OspfAuthProfileMd5",
    )

    def _setup(self):
        self._xpaths.add_profile(value='/auth-profile')

        params = []

        params.append(VersionedParamPath(
            'type', values=['password', 'md5'], path='{type}'))
        params.append(VersionedParamPath(
            'password', condition={'type': 'password'}, path='{type}'))

        self._params = tuple(params)


class OspfAuthProfileMd5(VersionedPanObject):
    """OSPF Authentication Profile

    Args:
        keyid (int): Identifier for key
        key (str): The authentication key
        preferred (bool): This key is preferred

    """
    SUFFIX = ENTRY
    NAME = 'keyid'

    def _setup(self):
        self._xpaths.add_profile(value='/md5')

        params = []

        params.append(VersionedParamPath(
            'key', vartype='encrypted'))
        params.append(VersionedParamPath(
            'preferred', vartype='yesno'))

        self._params = tuple(params)


class OspfExportRules(VersionedPanObject):
    """OSPF Export Rules

    Args:
        name (str): IP subnet or :class:`pandevice.network.RedistributionProfile`
        new_path_type (str): New path type, 'ext-1' or 'ext-2' (Default: ext-2)
        new_tag (str): New tag (int or IP format)
        metric (int): Metric

    """
    SUFFIX = ENTRY

    def _setup(self):
        self._xpaths.add_profile(value='/export-rules')

        params = []

        params.append(VersionedParamPath(
            'new_path_type', default='ext-2', values=['ext-1', 'ext-2']))
        params.append(VersionedParamPath(
            'new_tag'))
        params.append(VersionedParamPath(
            'metric', vartype='int'))

        self._params = tuple(params)


class ManagementProfile(VersionedPanObject):
    """Interface management provile.

    Add to any of the following interfaces:

    * Layer3Subinterface
    * EthernetInterface
    * AggregateInterface
    * VlanInterface
    * LoopbackInterface
    * TunnelInterface

    Args:
        ping (bool): Enable ping
        telnet (bool): Enable telnet
        ssh (bool): Enable ssh
        http (bool): Enable http
        http_ocsp (bool): Enable http-ocsp
        https (bool): Enable https
        snmp (bool): Enable snmp
        response_pages (bool): Enable response pages
        userid_service (bool): Enable userid service
        userid_syslog_listener_ssl (bool): Enable userid syslog listener ssl
        userid_syslog_listener_udp (bool): Enable userid syslog listener udp
        permitted_ip (list): The list of permitted IP addresses

    """
    ROOT = Root.DEVICE
    SUFFIX = ENTRY

    def _setup(self):
        # xpaths
        self._xpaths.add_profile(
            value='/network/profiles/interface-management-profile')

        # params
        params = []

        yesnos = ('ping', 'telnet', 'ssh', 'http', 'http-ocsp', 'https',
                  'snmp', 'response-pages', 'userid-service',
                  'userid-syslog-listener-ssl', 'userid-syslog-listener-udp')
        for yn in yesnos:
            params.append(VersionedParamPath(
                yn, path=yn, vartype='yesno'))
        params.append(VersionedParamPath(
            'permitted-ip', path='permitted-ip', vartype='entry'))

        self._params = tuple(params)



# Copyright (c) 2014, Palo Alto Networks
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
try:
    from unittest import mock
except ImportError:
    import mock
import unittest
import uuid
import xml.etree.ElementTree as ET

import pan.xapi
import pandevice.base as Base
import pandevice.errors as Err


OBJECT_NAME = 'MyObjectName'
VSYS = 'vsys1'


class TestPanObject(unittest.TestCase):
    def setUp(self):
        self.obj = Base.PanObject(OBJECT_NAME)

    def verify_object(self, obj, **kwargs):
        self.assertEqual(
            kwargs.get('name', None),
            obj.name)
        self.assertEqual(
            kwargs.get('children', []),
            obj.children)
        self.assertEqual(
            kwargs.get('parent', None),
            obj.parent)

    def test_create_with_name(self):
        self.obj = Base.PanObject(OBJECT_NAME)

        self.verify_object(self.obj, name=OBJECT_NAME)

    def test_create_without_name(self):
        self.obj = Base.PanObject()

        self.verify_object(self.obj)

    def test_str_of_object_with_name(self):
        self.assertEqual(OBJECT_NAME, str(self.obj))

    def test_str_of_object_without_name(self):
        self.obj = Base.PanObject()

        self.assertEqual('None', str(self.obj))

    def test_has_callable_variables(self):
        self.assertTrue(callable(self.obj.variables))

    def test_property_vsys_without_parent(self):
        self.assertIsNone(self.obj.vsys)

    def test_property_vsys_with_parent(self):
        self.obj.parent = mock.Mock(vsys=VSYS)

        self.assertEqual(VSYS, self.obj.vsys)

    def test_property_vsys_raises_error(self):
        self.assertRaises(
            Err.PanDeviceError,
            setattr,
            self.obj, 'vsys', 'foo')

    def test_property_uid(self):
        expected = OBJECT_NAME

        ret_val = self.obj.uid

        self.assertEqual(expected, ret_val)

    def test_add_without_children(self):
        CHILD_NAME = 'child'
        child = Base.PanObject(CHILD_NAME)

        ret_value = self.obj.add(child)

        self.assertEqual(child, ret_value)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=[child, ])
        self.verify_object(
            child,
            name=CHILD_NAME,
            parent=self.obj)

    def test_add_with_children(self):
        CHILD1_NAME = 'FirstChild'
        child1 = Base.PanObject(CHILD1_NAME)
        child1.parent = self.obj

        self.obj.children = [child1, ]

        CHILD2_NAME = 'SecondChild'
        child2 = Base.PanObject(CHILD2_NAME)

        ret_val = self.obj.add(child2)

        self.assertEqual(child2, ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=[child1, child2])
        self.verify_object(
            child1,
            name=CHILD1_NAME,
            parent=self.obj)
        self.verify_object(
            child2,
            name=CHILD2_NAME,
            parent=self.obj)

    def test_insert_without_children(self):
        CHILD_NAME = 'Child'
        child = Base.PanObject(CHILD_NAME)

        ret_val = self.obj.insert(0, child)

        self.assertEqual(child, ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=[child, ])
        self.verify_object(
            child,
            name=CHILD_NAME,
            parent=self.obj)

    def test_insert_with_children(self):
        CHILD1_NAME = 'FirstChild'
        child1 = Base.PanObject(CHILD1_NAME)
        child1.parent = self.obj

        CHILD3_NAME = 'ThirdChild'
        child3 = Base.PanObject(CHILD3_NAME)
        child3.parent = self.obj

        self.obj.children = [child1, child3]

        CHILD2_NAME = 'SecondChild'
        child2 = Base.PanObject(CHILD2_NAME)

        ret_val = self.obj.insert(1, child2)

        self.assertEqual(child2, ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=[child1, child2, child3])
        self.verify_object(
            child1,
            name=CHILD1_NAME,
            parent=self.obj)
        self.verify_object(
            child2,
            name=CHILD2_NAME,
            parent=self.obj)
        self.verify_object(
            child3,
            name=CHILD3_NAME,
            parent=self.obj)

    def test_extend_without_children(self):
        CHILD1_NAME = 'FirstChild'
        child1 = Base.PanObject(CHILD1_NAME)

        CHILD2_NAME = 'SecondChild'
        child2 = Base.PanObject(CHILD2_NAME)

        children = [child1, child2]

        ret_val = self.obj.extend(children)

        self.assertIsNone(ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=children)
        self.verify_object(
            child1,
            name=CHILD1_NAME,
            parent=self.obj)
        self.verify_object(
            child2,
            name=CHILD2_NAME,
            parent=self.obj)

    def test_extend_with_children(self):
        CHILD1_NAME = 'FirstChild'
        child1 = Base.PanObject(CHILD1_NAME)
        child1.parent = self.obj

        self.obj.children = [child1, ]

        CHILD2_NAME = 'SecondChild'
        child2 = Base.PanObject(CHILD2_NAME)

        CHILD3_NAME = 'ThirdChild'
        child3 = Base.PanObject(CHILD3_NAME)

        new_children = [child2, child3]
        all_children = [child1, child2, child3]

        ret_val = self.obj.extend(new_children)

        self.assertIsNone(ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=all_children)
        self.verify_object(
            child1,
            name=CHILD1_NAME,
            parent=self.obj)
        self.verify_object(
            child2,
            name=CHILD2_NAME,
            parent=self.obj)
        self.verify_object(
            child3,
            name=CHILD3_NAME,
            parent=self.obj)

    def test_pop(self):
        CHILD_NAME = 'Child'
        child = Base.PanObject(CHILD_NAME)
        child.parent = self.obj

        self.obj.children = [child, ]

        ret_val = self.obj.pop(0)

        self.assertEqual(child, ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME)
        self.verify_object(
            child,
            name=CHILD_NAME)

    def test_pop_raises_error(self):
        '''An invalid index should raise IndexError.'''
        self.assertRaises(
            IndexError,
            self.obj.pop, 0)

    def test_remove(self):
        CHILD1_NAME = 'Child1'
        child1 = Base.PanObject(CHILD1_NAME)
        child1.parent = self.obj

        CHILD2_NAME = 'Child2'
        child2 = Base.PanObject(CHILD2_NAME)
        child2.parent = self.obj

        self.obj.children = [child1, child2]

        ret_val = self.obj.remove(child2)

        self.assertIsNone(ret_val)
        self.verify_object(
            self.obj,
            name=OBJECT_NAME,
            children=[child1, ])
        self.verify_object(
            child1,
            name=CHILD1_NAME,
            parent=self.obj)
        self.verify_object(
            child2,
            name=CHILD2_NAME)

    def test_remove_raises_error(self):
        '''An invalid child should raise ValueError.'''
        CHILD1_NAME = 'Child1'
        child1 = Base.PanObject(CHILD1_NAME)
        child1.parent = self.obj

        CHILD2_NAME = 'Child2'
        child2 = Base.PanObject(CHILD2_NAME)

        self.obj.children = [child1, ]

        self.assertRaises(
            ValueError,
            self.obj.remove, child2)

    def test_remove_by_name_when_find_returns_index(self):
        CHILD_NAME = 'MyChild'
        self.obj.children = [1, 2, 3]

        INDEX_VALUE = 4
        self.obj.find_index = mock.Mock(return_value=INDEX_VALUE)

        POP_RETURN_VALUE = 'foo'
        self.obj.pop = mock.Mock(return_value=POP_RETURN_VALUE)

        ret_val = self.obj.remove_by_name(CHILD_NAME, None)

        self.assertEqual(POP_RETURN_VALUE, ret_val)
        self.obj.find_index.assert_called_once_with(CHILD_NAME, None)
        self.obj.pop.assert_called_once_with(INDEX_VALUE)

    def test_remove_by_name_when_find_returns_none(self):
        CHILD_NAME = 'foo'
        self.obj.children = ['a', 'b', 'c']

        self.obj.find_index = mock.Mock(return_value=None)

        ret_val = self.obj.remove_by_name(CHILD_NAME, None)

        self.assertIsNone(ret_val)
        self.obj.find_index.assert_called_once_with(CHILD_NAME, None)

    # Skipping removeall

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_xpath_with_entry_suffix(self, m_uid):
        Uid = 'baz'
        ParentValue = '/foo'
        SelfXpath = '/bar'
        Suffix = Base.ENTRY
        expected = ParentValue + SelfXpath + Base.ENTRY % (Uid, )

        m_uid.return_value = Uid
        self.obj._parent_xpath = mock.Mock(
            return_value=ParentValue)
        self.obj.XPATH = SelfXpath
        self.obj.SUFFIX = Suffix

        ret_val = self.obj.xpath()

        self.assertEqual(expected, ret_val)
        self.obj._parent_xpath.assert_called_once_with()

    # Skipping xpath_nosuffix

    # Skipping xpath_short

    def test_parent_xpath_for_none_parent(self):
        '''None parent results in empty xpath string.'''
        ret_val = self.obj._parent_xpath()

        self.assertEqual('', ret_val)

    def test_parent_xpath_for_xpath_parent(self):
        '''Default case for parent xpath.'''
        Path = '/hello/world'

        spec = {
            '_build_xpath.return_value': Path,
        }
        self.obj.parent = mock.Mock(spec=['_build_xpath', ], **spec)

        ret_val = self.obj._parent_xpath()

        self.assertEqual(Path, ret_val)
        self.obj.parent._build_xpath.assert_called_once_with(0)

    def test_xpath_vsys_without_parent(self):
        ret_val = self.obj.xpath_vsys()

        self.assertIsNone(ret_val)

    def test_xpath_vsys_with_parent(self):
        expected_value = 'foo'
        spec = {
            'xpath_vsys.return_value': expected_value,
        }
        self.obj.parent = mock.Mock(**spec)

        ret_val = self.obj.xpath_vsys()

        self.assertEqual(expected_value, ret_val)
        self.obj.parent.xpath_vsys.assert_called_once_with()

    def test_xpath_panorama_without_parent(self):
        ret_val = self.obj.xpath_panorama()

        self.assertIsNone(ret_val)

    def test_xpath_panorama_with_parent(self):
        expected_value = 'foo'
        spec = {
            'xpath_panorama.return_value': expected_value,
        }
        self.obj.parent = mock.Mock(**spec)

        ret_val = self.obj.xpath_panorama()

        self.assertEqual(expected_value, ret_val)
        self.obj.parent.xpath_panorama.assert_called_once_with()

    # Skip element()

    @mock.patch('pandevice.base.ET')
    def test_element_str(self, m_ET):
        Element_Value = 42
        self.obj.element = mock.Mock(return_value=Element_Value)

        Tostring_Value = '42'
        spec = {
            'tostring.return_value': Tostring_Value,
        }
        m_ET.configure_mock(**spec)

        ret_val = self.obj.element_str()

        self.assertEqual(Tostring_Value, ret_val)
        self.obj.element.assert_called_once_with()
        m_ET.tostring.assert_called_once_with(Element_Value, encoding='utf-8')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    @mock.patch('pandevice.base.ET')
    def test_root_element_with_entry_suffix(self, m_ET, m_uid):
        self.obj.SUFFIX = Base.ENTRY
        Uid = 'uid'
        expected = 'Value'

        spec = {
            'Element.return_value': expected,
        }
        m_ET.configure_mock(**spec)
        m_uid.return_value = Uid

        ret_val = self.obj._root_element()

        self.assertEqual(expected, ret_val)
        m_ET.Element.assert_called_once_with('entry', {'name': Uid})

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    @mock.patch('pandevice.base.ET')
    def test_root_element_with_member_suffix(self, m_ET, m_uid):
        self.obj.SUFFIX = Base.MEMBER
        Uid = 'uid'
        expected = mock.Mock(text=Uid)

        spec = {
            'Element.return_value': mock.Mock(),
        }
        m_ET.configure_mock(**spec)
        m_uid.return_value = Uid

        ret_val = self.obj._root_element()

        self.assertEqual(Uid, ret_val.text)
        m_ET.Element.assert_called_once_with('member')

    @mock.patch('pandevice.base.ET')
    def test_root_element_with_none_suffix_no_slashes(self, m_ET):
        self.obj.SUFFIX = None

        expected_tag = 'baz'
        full_path = expected_tag
        self.obj.XPATH = full_path

        expected_value = '42'
        spec = {
            'Element.return_value': expected_value,
        }
        m_ET.configure_mock(**spec)

        ret_val = self.obj._root_element()

        self.assertEqual(expected_value, ret_val)
        m_ET.Element.assert_called_once_with(expected_tag)

    @mock.patch('pandevice.base.ET')
    def test_root_element_with_none_suffix_multiple_slashes(self, m_ET):
        self.obj.SUFFIX = None

        expected_tag = 'baz'
        full_path = '/foo/bar/baz'
        self.obj.XPATH = full_path

        expected_value = '42'
        spec = {
            'Element.return_value': expected_value,
        }
        m_ET.configure_mock(**spec)

        ret_val = self.obj._root_element()

        self.assertEqual(expected_value, ret_val)
        m_ET.Element.assert_called_once_with(expected_tag)

    # Skip _subelements

    def test_check_child_methods_for_name_not_in_childmethods(self):
        spec = {
            '_check_child_methods.return_value': None,
        }
        for x in range(3):
            m = mock.Mock(**spec)
            self.obj.children.append(m)

        Method = str(uuid.uuid4()).replace('-', '_')

        ret_val = self.obj._check_child_methods(Method)

        self.assertIsNone(ret_val)
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with(Method)

    def test_check_child_methods_for_name_in_childmethods(self):
        spec = {
            '_check_child_methods.return_value': None,
        }
        for x in range(3):
            m = mock.Mock(**spec)
            self.obj.children.append(m)

        Method = str(uuid.uuid4()).replace('-', '_')
        self.obj.CHILDMETHODS += (Method, )
        setattr(self.obj, 'child_{0}'.format(Method), mock.Mock())

        ret_val = self.obj._check_child_methods(Method)

        self.assertIsNone(ret_val)
        m = getattr(self.obj, 'child_{0}'.format(Method))
        m.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with(Method)

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_apply_with_ha_sync(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'
        PanDeviceElementStr = 'element string'

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=PanDeviceXpath)
        self.obj.element_str = mock.Mock(return_value=PanDeviceElementStr)
        m_uid.return_value = 'uid'
        for x in range(3):
            child = mock.Mock(**spec)
            self.obj.children.append(child)

        ret_val = self.obj.apply()

        self.assertIsNone(ret_val)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.active().xapi.edit.assert_called_once_with(
            PanDeviceXpath,
            PanDeviceElementStr,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath.assert_called_once_with()
        self.obj.element_str.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('apply')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_apply_without_ha_sync(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'
        PanDeviceElementStr = 'element string'

        self.obj.HA_SYNC = False

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=PanDeviceXpath)
        self.obj.element_str = mock.Mock(return_value=PanDeviceElementStr)
        m_uid.return_value = 'uid'
        for x in range(3):
            child = mock.Mock(**spec)
            self.obj.children.append(child)

        ret_val = self.obj.apply()

        self.assertIsNone(ret_val)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.xapi.edit.assert_called_once_with(
            PanDeviceXpath,
            PanDeviceElementStr,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath.assert_called_once_with()
        self.obj.element_str.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('apply')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_create_with_ha_sync(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'
        PanDeviceElementStr = 'element string'

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath_short = mock.Mock(return_value=PanDeviceXpath)
        self.obj.element_str = mock.Mock(return_value=PanDeviceElementStr)
        m_uid.return_value = 'uid'
        for x in range(3):
            child = mock.Mock(**spec)
            self.obj.children.append(child)

        ret_val = self.obj.create()

        self.assertIsNone(ret_val)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.active().xapi.set.assert_called_once_with(
            PanDeviceXpath,
            PanDeviceElementStr,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath_short.assert_called_once_with()
        self.obj.element_str.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('create')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_create_without_ha_sync(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'
        PanDeviceElementStr = 'element string'

        self.obj.HA_SYNC = False

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath_short = mock.Mock(return_value=PanDeviceXpath)
        self.obj.element_str = mock.Mock(return_value=PanDeviceElementStr)
        m_uid.return_value = 'uid'
        for x in range(3):
            child = mock.Mock()
            self.obj.children.append(child)

        ret_val = self.obj.create()

        self.assertIsNone(ret_val)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.xapi.set.assert_called_once_with(
            PanDeviceXpath,
            PanDeviceElementStr,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath_short.assert_called_once_with()
        self.obj.element_str.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('create')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_delete_with_ha_sync_no_parent(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=PanDeviceXpath)
        m_uid.return_value = 'uid'
        for x in range(3):
            child = mock.Mock(**spec)
            self.obj.children.append(child)

        ret_val = self.obj.delete()

        self.assertIsNone(ret_val)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.active().xapi.delete.assert_called_once_with(
            PanDeviceXpath,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('delete')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_delete_with_ha_sync_and_parent(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'
        Uid = 'uid'

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.parent = mock.Mock()
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=PanDeviceXpath)
        m_uid.return_value = Uid
        for x in range(3):
            child = mock.Mock(**spec)
            self.obj.children.append(child)

        ret_val = self.obj.delete()

        self.assertIsNone(ret_val)
        self.obj.parent.remove.assert_called_once_with(self.obj)
        m_pandevice.set_config_changed.assert_called_once_with()
        m_pandevice.active().xapi.delete.assert_called_once_with(
            PanDeviceXpath,
            retry_on_peer=self.obj.HA_SYNC,
        )
        self.obj.xpath.assert_called_once_with()
        for c in self.obj.children:
            c._check_child_methods.assert_called_once_with('delete')

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test_delete_without_ha_sync(self, m_uid):
        PanDeviceId = '42'
        PanDeviceXpath = 'path'

        m_uid.return_value = 'uid'

        self.obj.HA_SYNC = False

        spec = {
            'id': PanDeviceId,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=PanDeviceXpath)
        for x in range(3):
            child = mock.Mock()

    # Skip update

    # Skip refresh

    # Skip refresh_variable

    # Skip _refresh_children

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_default_args_none_suffix(self, m_uid):
        Xpath = '/x/path'
        lasttag = ''

        expected = 'foo'
        spec = {
            'find.return_value': expected,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)

        ret_val = self.obj._refresh_xml(False, True)

        self.assertEqual(expected, ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_default_args_with_member_suffix(self, m_uid):
        Xpath = '/x/path'
        lasttag = 'member'

        expected = 'foo'
        spec = {
            'find.return_value': expected,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        self.obj.SUFFIX = Base.MEMBER

        ret_val = self.obj._refresh_xml(False, True)

        self.assertEqual(expected, ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_default_args_with_entry_suffix(self, m_uid):
        Xpath = '/x/path'
        lasttag = 'entry'

        expected = 'foo'
        spec = {
            'find.return_value': expected,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        self.obj.SUFFIX = Base.ENTRY

        ret_val = self.obj._refresh_xml(False, True)

        self.assertEqual(expected, ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_with_running_config(self, m_uid):
        Xpath = '/x/path'
        lasttag = ''

        expected = 'foo'
        spec = {
            'find.return_value': expected,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.show.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        self.obj.refresh = mock.Mock()

        ret_val = self.obj._refresh_xml(True, True)

        self.assertEqual(expected, ret_val)
        m_pandevice.xapi.show.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_no_refresh_children(self, m_uid):
        Xpath = '/x/path'
        lasttag = ''

        expected = 'foo'
        spec = {
            'find.return_value': expected,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        self.obj.refresh = mock.Mock()

        ret_val = self.obj._refresh_xml(False, False)

        self.assertEqual(expected, ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_api_action_raises_pannosuchnode_with_exceptions_on_raises_error(self, m_uid):
        Xpath = '/x/path'

        spec = {
            'id': 'myid',
            'xapi.get.side_effect': Err.PanNoSuchNode,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        m_uid.return_value = 'uid'

        self.assertRaises(
            Err.PanObjectMissing,
            self.obj._refresh_xml, False, True)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_api_action_raises_pannosuchnode_with_exceptions_off_returns_none(self, m_uid):
        Xpath = '/x/path'

        spec = {
            'id': 'myid',
            'xapi.get.side_effect': Err.PanNoSuchNode,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        m_uid.return_value = 'uid'

        ret_val = self.obj._refresh_xml(False, False)

        self.assertIsNone(ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_api_action_raises_panxapierror_with_exceptions_on_raises_error(self, m_uid):
        Xpath = '/x/path'

        spec = {
            'id': 'myid',
            'xapi.get.side_effect': pan.xapi.PanXapiError,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        m_uid.return_value = 'uid'

        self.assertRaises(
            Err.PanObjectMissing,
            self.obj._refresh_xml, False, True)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_api_action_raises_panxapierror_with_exceptions_off_returns_none(self, m_uid):
        Xpath = '/x/path'

        spec = {
            'id': 'myid',
            'xapi.get.side_effect': pan.xapi.PanXapiError,
        }
        m_pandevice = mock.Mock(**spec)
        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)
        m_uid.return_value = 'uid'

        ret_val = self.obj._refresh_xml(False, False)

        self.assertIsNone(ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_find_fails_with_exceptions_on_raises_error(self, m_uid):
        Xpath = '/x/path'
        lasttag = ''

        expected = 'foo'
        spec = {
            'find.return_value': None,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)

        self.assertRaises(
            Err.PanObjectMissing,
            self.obj._refresh_xml, False, True)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    @mock.patch('pandevice.base.PanObject.uid', new_callable=mock.PropertyMock)
    def test__refresh_xml_find_fails_with_exceptions_off_returns_none(self, m_uid):
        '''Requires exceptions=False.'''
        Xpath = '/x/path'
        lasttag = ''

        expected = 'foo'
        spec = {
            'find.return_value': None,
        }
        m_root = mock.Mock(**spec)
        m_uid.return_value = 'uid'

        spec = {
            'id': 'myid',
            'xapi.get.return_value': m_root,
        }
        m_pandevice = mock.Mock(**spec)

        self.obj.nearest_pandevice = mock.Mock(return_value=m_pandevice)
        self.obj.xpath = mock.Mock(return_value=Xpath)

        ret_val = self.obj._refresh_xml(False, False)

        self.assertIsNone(ret_val)
        m_pandevice.xapi.get.assert_called_once_with(
            Xpath,
            retry_on_peer=self.obj.HA_SYNC)
        self.obj.xpath.assert_called_once_with()
        m_root.find.assert_called_once_with(
            'result/{0}'.format(lasttag))

    def test_nearest_pandevice(self):
        expected = 'return value'

        self.obj._nearest_pandevice = mock.Mock(return_value=expected)

        ret_val = self.obj.nearest_pandevice()

        self.assertEqual(expected, ret_val)
        self.obj._nearest_pandevice.assert_called_once_with()

    def test__nearest_pandevice_with_parent(self):
        expected = 'ParentObject'

        spec = {
            '_nearest_pandevice.return_value': expected,
        }
        self.obj.parent = mock.Mock(**spec)

        ret_val = self.obj._nearest_pandevice()

        self.assertEqual(expected, ret_val)
        self.obj.parent._nearest_pandevice.assert_called_once_with()

    def test__nearest_pandevice_without_parent_raises_error(self):
        self.assertRaises(
            Err.PanDeviceNotSet,
            self.obj._nearest_pandevice)

    def test_panorama_with_parent(self):
        expected = 'PanoramaObject'

        spec = {
            'panorama.return_value': expected,
        }
        self.obj.parent = mock.Mock(**spec)

        ret_val = self.obj.panorama()

        self.assertEqual(expected, ret_val)
        self.obj.parent.panorama.assert_called_once_with()

    def test_panorama_without_parent_raises_error(self):
        self.assertRaises(
            Err.PanDeviceNotSet,
            self.obj.panorama)

    def test_devicegroup_with_parent(self):
        expected = 'DeviceGroup'

        spec = {
            'devicegroup.return_value': expected,
        }
        self.obj.parent = mock.Mock(**spec)

        ret_val = self.obj.devicegroup()

        self.assertEqual(expected, ret_val)
        self.obj.parent.devicegroup.assert_called_once_with()

    def test_devicegroup_without_parent(self):
        ret_val = self.obj.devicegroup()

        self.assertIsNone(ret_val)

    # Skip find

    # Skip findall

    # Skip find_or_create

    # Skip findall_or_create

    # Skip find_index

    # Skip applyall

    # Skip refreshall

    # Skip refreshall_from_xml

    # Skip _parse_xml

class TestParamPath(unittest.TestCase):
    def setUp(self):
        self.elm = ET.Element('myroot')

    def test_element_for_exclude_returns_none(self):
        settings = {'baz': 'jack'}
        p = Base.ParamPath('baz', path='foo/bar', vartype=None,
                           condition=None, values=None, exclude=True)

        result = p.element(self.elm, settings, False)

        self.assertIsNone(result)

    def test_element_path_has_variable(self):
        p = Base.ParamPath('baz', path='{mode}/bar/baz', vartype=None,
                           condition=None, values=None)
        settings = {'baz': 'jack', 'mode': 'layer3'}

        result = p.element(self.elm, settings, False)
        self.assertIsNotNone(result)

        elm = result.find('./layer3/bar/baz')
        self.assertIsNotNone(elm, msg='Failed: elm = {0}'.format(ET.tostring(result)))
        self.assertEqual(settings['baz'], elm.text)

    def test_element_for_vartype_member_for_string(self):
        p = Base.ParamPath('baz', path='foo/bar/baz', vartype='member',
                           condition=None, values=None)
        settings = {'baz': 'jack'}

        result = p.element(self.elm, settings, False)
        self.assertIsNotNone(result)

        elm = result.findall('./foo/bar/baz/member')
        self.assertTrue(elm)
        self.assertEqual(1, len(elm))
        self.assertEqual(settings['baz'], elm[0].text)

    def test_element_for_vartype_member_for_list(self):
        p = Base.ParamPath('baz', path='foo/bar/baz', vartype='member',
                           condition=None, values=None)
        settings = {'baz': ['jack', 'john', 'jane', 'margret']}

        result = p.element(self.elm, settings, False)
        self.assertIsNotNone(result)

        elms = result.findall('./foo/bar/baz/member')
        self.assertEqual(len(settings['baz']), len(elms))

        for elm in elms:
            self.assertTrue(elm.text in settings['baz'])


class Abouter(object):
    def __init__(self, mode='layer3'):
        self.mode = mode

    def _about_object(self):
        return {'mode': self.mode}

class ParentClass1(Abouter):
    pass

class ParentClass2(Abouter):
    pass

class UnassociatedParent(Abouter):
    pass


class TestParentAwareXpathBasics(unittest.TestCase):
    DEFAULT_PATH_1 = '/default/path/1'
    DEFAULT_PATH_2 = '/default/path/2'
    SPECIFIED_PATH_1 = '/some/specific/path/1'
    SPECIFIED_PATH_2 = '/some/specific/path/2'

    def setUp(self):
        self.obj = Base.ParentAwareXpath()
        self.obj.add_profile(value=self.DEFAULT_PATH_1)
        self.obj.add_profile('1.0.0', self.DEFAULT_PATH_2)
        self.obj.add_profile(value=self.SPECIFIED_PATH_1,
                             parents=('ParentClass1', 'ParentClass2'))
        self.obj.add_profile('2.0.0', self.SPECIFIED_PATH_2,
                             ('ParentClass1', 'ParentClass2'))

    def test_old_default_xpath(self):
        parent = UnassociatedParent()

        self.assertEqual(
            self.DEFAULT_PATH_1,
            self.obj._get_versioned_value(
                (0, 5, 0), parent))

    def test_new_default_xpath(self):
        parent = UnassociatedParent()

        self.assertEqual(
            self.DEFAULT_PATH_2,
            self.obj._get_versioned_value(
                (1, 0, 0), parent))

    def test_old_specefied_xpath_for_class1(self):
        parent = ParentClass1()

        self.assertEqual(
            self.SPECIFIED_PATH_1,
            self.obj._get_versioned_value(
                (0, 5, 0), parent))

    def test_new_specefied_xpath_for_class1(self):
        parent = ParentClass1()

        self.assertEqual(
            self.SPECIFIED_PATH_2,
            self.obj._get_versioned_value(
                (2, 0, 0), parent))

    def test_old_specefied_xpath_for_class2(self):
        parent = ParentClass2()

        self.assertEqual(
            self.SPECIFIED_PATH_1,
            self.obj._get_versioned_value(
                (0, 0, 0), parent))

    def test_new_specefied_xpath_for_class2(self):
        parent = ParentClass2()

        self.assertEqual(
            self.SPECIFIED_PATH_2,
            self.obj._get_versioned_value(
                (5, 0, 0), parent))

    def test_no_parent_gets_newest_version(self):
        parent = None

        self.assertEqual(
            self.DEFAULT_PATH_2,
            self.obj._get_versioned_value(
                Base.VersionedPanObject._UNKNOWN_PANOS_VERSION, parent))

    def test_no_fallback_raises_value_error(self):
        parent = None
        obj = Base.ParentAwareXpath()
        obj.add_profile(
            parents=('ParentClass1', ),
            value='/some/path',
        )

        self.assertRaises(
            ValueError,
            obj._get_versioned_value,
            (1, 0, 0), parent)


class TestParentAwareXpathWithParams(unittest.TestCase):
    OLD_LAYER3_PATH = '/units/layer3/old'
    NEW_LAYER3_PATH = '/units/layer3/new'
    OLD_LAYER2_PATH = '/units/layer2/old'
    NEW_LAYER2_PATH = '/units/layer2/new'

    def setUp(self):
        self.obj = Base.ParentAwareXpath()
        self.obj.add_profile(
            parents=('ParentClass1', None),
            value=self.OLD_LAYER3_PATH)
        self.obj.add_profile(
            version='1.0.0',
            parents=('ParentClass1', None),
            value=self.NEW_LAYER3_PATH)
        self.obj.add_profile(
            parents=('ParentClass1', ),
            parent_param='mode',
            parent_param_values=['junk', 'layer2'],
            value=self.OLD_LAYER2_PATH)
        self.obj.add_profile(
            version='2.0.0',
            parents=('ParentClass1', ),
            parent_param='mode',
            parent_param_values=['junk', 'layer2'],
            value=self.NEW_LAYER2_PATH)

    def test_old_default_path(self):
        parent = UnassociatedParent('foo')

        self.assertEqual(
            self.OLD_LAYER3_PATH,
            self.obj._get_versioned_value(
                (0, 5, 0), parent))

    def test_known_parent_and_param_for_old_l3_path(self):
        parent = ParentClass1()

        self.assertEqual(
            self.OLD_LAYER3_PATH,
            self.obj._get_versioned_value(
                (0, 5, 0), parent))

    def test_known_parent_and_param_for_new_l3_path(self):
        parent = ParentClass1()

        self.assertEqual(
            self.NEW_LAYER3_PATH,
            self.obj._get_versioned_value(
                (1, 5, 0), parent))

    def test_known_parent_and_param_for_old_l2_path(self):
        parent = ParentClass1('layer2')

        self.assertEqual(
            self.OLD_LAYER2_PATH,
            self.obj._get_versioned_value(
                (0, 1, 0), parent))

    def test_known_parent_and_param_for_new_l2_path(self):
        parent = ParentClass1('layer2')

        self.assertEqual(
            self.NEW_LAYER2_PATH,
            self.obj._get_versioned_value(
                (5, 1, 0), parent))

    def test_no_parent_gets_newest_default(self):
        parent = None

        self.assertEqual(
            self.NEW_LAYER3_PATH,
            self.obj._get_versioned_value(
                Base.VersionedPanObject._UNKNOWN_PANOS_VERSION, parent))


class MyVersionedObject(Base.VersionedPanObject):
    SUFFIX = Base.ENTRY

    def _setup(self):
        params = []

        params.append(Base.VersionedParamPath(
            'entries', path='multiple/entries', vartype='entry'))
        params.append(Base.VersionedParamPath(
            'members', path='multiple/members', vartype='member'))
        params.append(Base.VersionedParamPath(
            'someint', path='someint', vartype='int'))

        self._params = tuple(params)


class TestEqual(unittest.TestCase):
    def test_ordered(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)

        self.assertTrue(o1.equal(o2))

    def test_unordered_entries(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['b', 'a'], ['c', 'd'], 5)

        self.assertTrue(o1.equal(o2))

    def test_unordered_members(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['a', 'b'], ['d', 'c'], 5)

        self.assertTrue(o1.equal(o2))

    def test_values_are_unchanged_after_comparison(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['b', 'a'], ['d', 'c'], 5)

        o1.equal(o2)

        self.assertEqual(o1.entries, ['a', 'b'])
        self.assertEqual(o1.members, ['c', 'd'])
        self.assertEqual(o2.entries, ['b', 'a'])
        self.assertEqual(o2.members, ['d', 'c'])

    def test_str_list_field_is_equal(self):
        o1 = MyVersionedObject('a', ['a', ], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', 'a', ['c', 'd'], 5)

        self.assertTrue(o1.equal(o2))

    def test_unequal_entries_returns_false(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['a', 'i'], ['c', 'd'], 5)

        self.assertFalse(o1.equal(o2))

    def test_unequal_members_returns_false(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['a', 'b'], ['c', 'i'], 5)

        self.assertFalse(o1.equal(o2))

    def test_unequal_ints_returns_false(self):
        o1 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 5)
        o2 = MyVersionedObject('a', ['a', 'b'], ['c', 'd'], 6)

        self.assertFalse(o1.equal(o2))


if __name__=='__main__':
    unittest.main()
